import asyncio
import contextlib
import socket


async def check_internet_connection() -> bool:
    """
    Check internet connection by trying to access Google's website
    TODO: we may need to try a few times before reporting no internet access
    """
    def _check_connection() -> bool:
        with contextlib.suppress(OSError):
            conn = socket.create_connection(("www.google.com", 80))
            conn.close()

            return True

        return False

    loop = asyncio.get_event_loop()

    return await loop.run_in_executor(None, _check_connection)



import logging


def init_logger(log_level: str = "DEBUG", third_party_level: str = "WARNING") -> None:
    log_level = log_level.upper()
    third_party_level = third_party_level.upper()

    logging.basicConfig(
        format="[%(levelname)s] - %(asctime)s - %(name)s::%(funcName)s() - %(message)s",
        level=log_level,
    )

    logging.getLogger('aiosqlite').setLevel(third_party_level)
    logging.getLogger('sqlalchemy').setLevel(third_party_level)
    logging.getLogger('sqlalchemy.engine').setLevel(third_party_level)
    logging.getLogger('aiogram').setLevel(third_party_level)



from datetime import datetime

from sqlalchemy import JSON, Boolean, Column, DateTime, Integer, String

from hirsh.entities.base import BaseEntity, StrEnum


class LogStatus(StrEnum):
    UP = "up"
    DOWN = "down"


class Log(BaseEntity):
    __tablename__ = "outage_monitor_logs"

    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    created_at = Column(DateTime, default=datetime.utcnow)
    resource = Column(String(60))
    status = Column(String(60))
    processed = Column(Boolean, default=False)
    context = Column(JSON, nullable=True)



from collections import defaultdict
from datetime import datetime
from typing import Any, Optional, Sequence

from hirsh.entities import Event, Log, LogStatus, Resources


class EventFinder:
    """
    TODO: Finish the event finder
      https://github.com/roma-glushko/hirsh/issues/2
    """
    def __init__(self) -> None:
        ...

    def find(self, logs: Sequence[Log]) -> list[Event]:
        events: dict[Resources, list[Event]] = defaultdict(list)
        context: dict[Resources, Optional[dict[str, Any]]] = {}
        processed_ids: list[int] = []

        for log in logs:
            resource_type = Resources(log.resource)

            if resource_type == Resources.DAEMON:
                # detected a daemon downtime
                latest_log: Optional[datetime] = None

                for type, context in context.items():
                    # flush events
                    latest_log = max(latest_log, context["finished_at"]) if latest_log else context["finished_at"]
                    events[type].append(Event(**context))

                context.clear()
                processed_ids.append(log.id)

                events[Resources.DAEMON] = Event(
                    started_at=latest_log,
                    ended_at=log.created_at,
                    resource=Resources.DAEMON,
                    status="down",
                )

            if resource_type not in context:
                context[resource_type] = {
                    "resource": resource_type,
                    "status": LogStatus(log.status),
                    "started_at": log.created_at,
                    "finished_at": None,
                }
                processed_ids.append(log.id)
                continue

            current_context = context.get(resource_type)

            current_context["finished_at"] = log.created_at

            if current_context["status"] != log.status:
                # end of the interval
                events[resource_type].append(Event(**current_context))

        return events



import asyncio
import logging
import signal
import time
from typing import Sequence

from hirsh.services.monitors import DaemonMonitor, Monitor

logger = logging.getLogger(__name__)


async def schedule(monitor: Monitor) -> None:
    """
    Schedule a monitor to execute every X seconds
    """

    def _until_next(last: float) -> float:
        time_took = time.time() - last

        return monitor.check_every_secs - time_took

    while True:
        time_start = time.time()

        try:
            await monitor.check()
        except asyncio.CancelledError:
            break
        except Exception:
            monitor.logger.exception("Error executing monitor check")

        await asyncio.sleep(_until_next(last=time_start))


class Daemon:
    def __init__(self, lifecycle_monitor: DaemonMonitor, monitors: Sequence[Monitor]) -> None:
        self._lifecycle_monitor = lifecycle_monitor
        self._monitors = monitors
        self._monitor_tasks: list[asyncio.Task] = []
        self._stopping = False

    async def start(self) -> None:
        logger.info("Starting up hirsh daemon")
        await self._lifecycle_monitor.startup()

        event_loop = asyncio.get_running_loop()

        for monitor in self._monitors:
            self._monitor_tasks.append(
                asyncio.create_task(schedule(monitor)),
            )

        for sig in (signal.SIGTERM, signal.SIGINT, signal.SIGQUIT):
            event_loop.add_signal_handler(sig, self.stop)

        await asyncio.gather(*self._monitor_tasks, return_exceptions=True)

        self.stop()

    def stop(self) -> None:
        if self._stopping:
            return

        self._stopping = True
        logger.info("Shutting down hirsh daemon")

        # TODO: allow to run async shutdown() method
        # await self._lifecycle_monitor.shutdown()

        for task, _monitor in zip(self._monitor_tasks, self._monitors):
            task.cancel()

        self._monitor_tasks.clear()

        logger.info("Hirsh daemon has shut down successfully")



import logging
from asyncio import current_task
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import AsyncGenerator, Generator

from sqlalchemy.ext.asyncio import AsyncSession, async_scoped_session, create_async_engine
from sqlalchemy.orm import sessionmaker

from hirsh.entities import BaseEntity

logger = logging.getLogger(__name__)


class Database:
    """
    Low-level SQLite session manager
    """

    def __init__(self, database_uri: str) -> None:
        self._engine = create_async_engine(database_uri, echo=True)

        self._session_factory = async_scoped_session(
            sessionmaker(
                bind=self._engine,
                autocommit=False,
                autoflush=False,
                class_=AsyncSession,
            ),
            scopefunc=current_task,
        )

    async def create_database(self) -> None:
        async with self._engine.begin() as conn:
            await conn.run_sync(BaseEntity.metadata.create_all)

    @asynccontextmanager
    async def session(self) -> AsyncGenerator[AbstractAsyncContextManager[AsyncSession], None]:
        session: AsyncSession = self._session_factory()
        try:
            yield session
        except Exception:
            logger.exception("Session rollback because of exception")
            await session.rollback()
            raise
        finally:
            await session.close()

    async def close(self) -> None:
        await self._engine.dispose()


async def init_database(database_uri: str) -> Generator[Database, None, None]:
    database = Database(database_uri)
    await database.create_database()
    yield database
    await database.close()



from hirsh.entities.base import BaseEntity, Resources
from hirsh.entities.events import Event
from hirsh.entities.logs import Log, LogStatus

__all__ = ("BaseEntity", "Log", "Event", "Resources", "LogStatus")



from typing import Any, Optional

from hirsh.entities import Log, LogStatus, Resources


class LogRepository:
    """
    Manage monitor logs in the database
    """

    def __init__(self, session_factory) -> None:
        self.session_factory = session_factory

    async def add_log(self, resource: Resources, status: LogStatus, context: Optional[dict[str, Any]] = None) -> None:
        async with self.session_factory() as session:
            log = Log(
                resource=str(resource),
                status=str(status),
                processed=False,
                context=context,
            )

            session.add(log)

            await session.commit()



import logging
from pathlib import Path

from dependency_injector import containers, providers
from dependency_injector.providers import Configuration, Resource, Singleton

from hirsh.config import load_from_yaml
from hirsh.daemon import Daemon
from hirsh.logging import init_logger
from hirsh.repositories import Database, EventRepository, LogRepository, init_database
from hirsh.services.monitors import DaemonMonitor, NetworkMonitor
from hirsh.services.notifiers import TelegramNotifier

logger = logging.getLogger(__name__)


class Runtime(containers.DeclarativeContainer):
    config = Configuration()

    db: Resource[Database] = Resource(init_database, database_uri=config.database.uri)

    log_repository: Singleton[LogRepository] = Singleton(
        LogRepository,
        session_factory=db.provided.session
    )

    event_repository: Singleton[EventRepository] = Singleton(
        EventRepository,
        session_factory=db.provided.session
    )

    notifier: Singleton[TelegramNotifier] = Singleton(
        TelegramNotifier,
        bot_token=config.telegram.bot_token,
        channel_ids=config.telegram.channel_ids,
    )

    daemon_monitor: Singleton[DaemonMonitor] = Singleton(
        DaemonMonitor,
        log_repository=log_repository,
        notifier=notifier,
    )

    daemon: Singleton[Daemon] = Singleton(
        Daemon,
        lifecycle_monitor=daemon_monitor,
        monitors=providers.List(
            Singleton(
                NetworkMonitor,
                log_repository=log_repository,
                notifier=notifier,
                check_every_secs=config.monitoring.check_every_secs
            ),
        )
    )


def create_runtime(config_path: Path) -> Runtime:
    logger.debug("Initializing runtime container..")

    runtime = Runtime()

    config = load_from_yaml(config_path)
    runtime.config.from_pydantic(config)

    init_logger(log_level=runtime.config.logging.level())

    return runtime



from enum import Enum

from sqlalchemy.ext.declarative import declarative_base


class StrEnum(str, Enum):
    def __str__(self):
        return str(self.value)


class Resources(StrEnum):
    DAEMON = "daemon"
    NETWORK = "network"


BaseEntity = declarative_base()



import logging

import socket

from hirsh.entities import LogStatus, Resources
from hirsh.repositories import LogRepository
from hirsh.services.network import check_internet_connection
from hirsh.services.notifiers import Notifier


class Monitor:
    def __init__(self, check_every_secs: int) -> None:
        self.check_every_secs = check_every_secs
        self.logger = logging.getLogger(self.__class__.__name__)

    async def check(self) -> None:
        raise NotImplementedError()


class DaemonMonitor:
    """
    Special type of monitors that tracks daemon lifecycle
    """

    def __init__(self, log_repository: LogRepository, notifier: Notifier) -> None:
        self._log_repository = log_repository
        self._notifier = notifier

    def _hostname(self) -> str:
        return socket.gethostname()

    async def startup(self) -> None:
        await self._log_repository.add_log(
            resource=Resources.DAEMON,
            status=LogStatus.UP,
        )

        await self._notifier.notify(f"📟 Hirsh ({self._hostname()}) is starting up..")

    async def shutdown(self) -> None:
        await self._log_repository.add_log(
            resource=Resources.DAEMON,
            status=LogStatus.DOWN,
        )

        await self._notifier.notify(f"📟 Hirsh ({self._hostname()}) is shutting down..")


class NetworkMonitor(Monitor):
    def __init__(self, log_repository: LogRepository, notifier: Notifier, check_every_secs: int = 60) -> None:
        super().__init__(check_every_secs)

        self._log_repository = log_repository
        self._notifier = notifier

    async def check(self) -> None:
        self.logger.debug("Checking network connection")

        network_status: bool = await check_internet_connection()

        if network_status:
            await self._notifier.notify("📡 Network is working at the moment")

        await self._log_repository.add_log(
            resource=Resources.NETWORK,
            status=LogStatus.UP if network_status else LogStatus.DOWN,
        )


class ElectricityMonitor(Monitor):
    """
    TODO: For the phase 1 of the project, we will implement network monitor only.
        Electricity monitoring could require more effort and planning to implement

    """



from datetime import datetime

from sqlalchemy import JSON, Boolean, Column, DateTime, Integer, String

from hirsh.entities.base import BaseEntity


class Event(BaseEntity):
    __tablename__ = "outage_monitor_events"

    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    started_at = Column(DateTime, default=datetime.utcnow)
    ended_at = Column(DateTime)
    resource = Column(String(60))
    status = Column(String(60))
    notified = Column(Boolean, default=False)
    context = Column(JSON, nullable=True)



from hirsh.entities import Resources
from hirsh.entities.events import Event


class EventRepository:
    """
    Manage monitor events in the database
    """

    def __init__(self, session_factory) -> None:
        self.session_factory = session_factory

    async def add_event(self, resource: Resources) -> None:
        async with self.session_factory() as session:
            log = Event(
                resource=resource,
            )

            session.add(log)
            session.commit()



from hirsh.repositories.database import Database, init_database
from hirsh.repositories.events import EventRepository
from hirsh.repositories.logs import LogRepository

__all__ = ("Database", "init_database", "LogRepository", "EventRepository", )



import logging


from aiogram import Bot, types
from aiogram.utils.exceptions import NetworkError


class Notifier:
    async def notify(self, message: str) -> None:
        raise NotImplementedError()


class TelegramNotifier(Notifier):
    def __init__(self, bot_token: str, channel_ids: list[int]) -> None:
        self.logger = logging.getLogger(self.__class__.__name__)

        self._bot = Bot(token=bot_token, parse_mode=types.ParseMode.HTML)
        self._channel_ids = channel_ids

    async def notify(self, message: str) -> None:
        try:
            for channel_id in self._channel_ids:
                await self._bot.send_message(
                    chat_id=channel_id,
                    text=message,
                    disable_notification=False,
                )
        except NetworkError:
            # TODO: Deferred message sending on no connection
            #  https://github.com/roma-glushko/hirsh/issues/1
            self.logger.warning(
                "No internet connection to notify the targets",
                extra={
                    "notification": message,
                },
            )


import logging
from pathlib import Path
from typing import Any, Optional

import yaml
from pydantic import BaseSettings, Field

logger = logging.getLogger(__name__)


class TelegramConfig(BaseSettings):
    bot_token: str
    channel_ids: list[int]


class DatabaseConfig(BaseSettings):
    uri: str  # TODO: Create SQLiteDns type (see pydantic.PostgresDsn)


class LoggingConfig(BaseSettings):
    level: str = "INFO"


class MonitoringConfig(BaseSettings):
    check_every_secs: int = 60


class Config(BaseSettings):
    logging: LoggingConfig = Field(default_factory=LoggingConfig)
    database: DatabaseConfig = Field(default_factory=DatabaseConfig)
    telegram: Optional[TelegramConfig] = Field(None)
    monitoring: MonitoringConfig = Field(default_factory=MonitoringConfig)


class ConfigReadFailed(RuntimeError):
    """
    Occurs when it's impossible to read the given YAML config file
    """


def load_from_yaml(config_path: Path) -> Config:
    config_path = config_path.expanduser()

    logger.debug(f"Reading '{config_path}' config file")

    if not config_path.is_file():
        raise ConfigReadFailed(f"'{config_path}' config file doesn't exist or not readable")

    with open(config_path, "r") as file:
        try:
            raw_config: dict[str, Any] = yaml.safe_load(file)

            return Config.parse_obj(raw_config)
        except yaml.YAMLError as e:
            raise ConfigReadFailed(f"Failed to read '{config_path}' config file") from e



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hirsh', 'hirsh.entities', 'hirsh.repositories', 'hirsh.services']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'SQLAlchemy[asyncio,mypy]>=1.4.44,<2.0.0',
 'aiogram>=2.23.1,<3.0.0',
 'aiosqlite>=0.17.0,<0.18.0',
 'dependency-injector>=4.40.0,<5.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'typer>=0.7.0,<0.8.0',
 'uvloop>=0.17.0,<0.18.0']

setup_kwargs = {
    'name': 'hirsh',
    'version': '0.0.2',
    'description': 'Resilient monitoring system that detects utility outages in unreliable environments (e.g. IoT, RaspberryPi, etc)',
    'long_description': "# Hirsh\n\n📟 Resilient monitoring system that detects utility outages in unreliable environments (e.g. IoT, RaspberryPi, etc).\n\nHirsh is designed and optimized for:\n\n- 💪 residency, robustness and self-healing \n- 📟 running in resource-constrained IoT-like unstable environments\n\n## Maturity\n\nThe project is in early MVP state. \n\nIt's being actively tested using my RaspberryPi Zero 2W under the current unstable Ukrainian infrastructure conditions.\n\n## Setups\n\nHirsh can be executed in any IoT device or board computer that supports Linux-like OS and Python 3.9+.\n\n- [Only Supported] Basic: The basic setup includes just the device.\n    The device is plugged into the main electricity circuit/outlet along with a router that provides network connection for the device.\n- UPS: TBU\n\n## Monitors\n\nIn theory, you can track any utilities your home has (e.g. electricity, network, gas, water, etc.). \nHowever, in practice it's the easiest to track:\n\n- network connection [Only Supported]\n- electricity supply\n\n## How does it work?\n\nTBU\n\n## Notifications\n\n### Telegram\n\nThe primary way to notify you about outages is via [Telegram bot](https://core.telegram.org/bots).\nYou need to create [a new bot](https://t.me/BotFather) and add it to a group or a channel.\n\n## References\n\n### Similar Projects\n\n- https://github.com/fabytm/Outage-Detector\n- https://github.com/nestukh/rpi-powerfail\n- https://www.kc4rcr.com/power-outage-notification/\n- https://homediyelectronics.com/projects/raspberrypi/poweroffdelay/powerfail\n- https://projects-raspberry.com/power-outage-sensor/\n- https://raspberrypi.stackexchange.com/questions/13538/raspberry-pi-to-email-when-power-outage-occurs\n\n### Python + RPi\n\n- https://medium.com/geekculture/gpio-programming-on-the-raspberry-pi-python-libraries-e12af7e0a812\n\n### AsyncIO and RPi\n\n- https://github.com/PierreRust/apigpio\n- https://beenje.github.io/blog/posts/experimenting-with-asyncio-on-a-raspberry-pi/\n- https://www.digikey.bg/en/maker/projects/getting-started-with-asyncio-in-micropython-raspberry-pi-pico/110b4243a2f544b6af60411a85f0437c\n- https://docs.micropython.org/en/latest/library/uasyncio.html\n\n### Deployment\n\n- https://github.com/beenje/legomac",
    'author': 'Roman Glushko',
    'author_email': 'roman.glushko.m@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging
from asyncio import run as aiorun
from pathlib import Path

import typer
import uvloop

from hirsh.daemon import Daemon
from hirsh.runtime import create_runtime

logger = logging.getLogger(__name__)


def run(working_dir: Path = "~/.hirsh", config_filename: str = "config.yaml") -> None:
    """
    Hirsh: Be first to know about outages in your apartments
    """

    async def __run() -> None:
        runtime = create_runtime(config_path=working_dir / config_filename)

        logger.debug("Initializing resources..")
        await runtime.init_resources()

        daemon: Daemon = await runtime.daemon()
        await daemon.start()

        logger.debug("Shutting down resources..")
        await runtime.shutdown_resources()

    aiorun(__run())


if __name__ == "__main__":
    uvloop.install()
    typer.run(run)
